
import io

from anyserver.encoder import Encoder


class YamlEncoder(Encoder):
    mime = "application/yaml"
    ext = [".yaml"]

    def __init__(self, yaml):
        # Provide the yaml module (we have no assurances its installed / importable)
        self.yaml = yaml

    def encode(self, data):
        with io.StringIO() as buffer:
            self.yaml.dump(data, buffer, default_flow_style=False)
            return buffer.getvalue()

    def decode(self, data):
        with io.StringIO(initial_value=data) as stream:
            return self.yaml.safe_load(stream)



import json

from urllib import parse
from functools import partial

from http.server import HTTPServer, SimpleHTTPRequestHandler

from anyserver.router import WebRequest, WebResponse
from anyserver.servers.abstract import AbstractServer


class Request(WebRequest):
    # Wrap your request object into serializable object
    def __init__(self, ctx): super().__init__(
        verb=ctx.command,
        path=ctx.path,
        head=ctx.headers,
        body=self.body(ctx, ctx.headers),
        params=self.query(ctx),
    )

    def query(self, ctx):
        path_parts = ctx.path.split('?')
        if len(path_parts) > 1:
            return parse.parse_qs(path_parts[1])
        return {}    

    def body(self, ctx, headers):
        # Only try and parse the body for known methods (eg: POST, PUT)
        if not ctx.command in ["POST", "PUT", "PATCH"]:
            return None 
        
        # Parse the body according to the content type
        ctype = headers['content-type'] if 'content-type' in headers else ''
        length = int(ctx.headers.get('content-length'))
        match ctype:
            case 'application/json':
                input = ctx.rfile.read(length).decode('utf8')
                data = json.loads(input)
            case 'application/x-www-form-urlencoded':
                input = ctx.rfile.read(length).decode('utf8')
                form = parse.parse_qs(input, keep_blank_values=1)
                data = {}
                for key in form:
                    if len(form[key]) > 1:
                        data[key] = form[key]
                    elif len(form[key]) == 1:
                        data[key] = form[key][0]
            case _:
                message = 'Content type "%s" cannot be parsed into a body.' % ctype
                raise Exception(message)

        return data


class Response(WebResponse):
    __sent = False

    # Wrap your response object into serializable object
    def __init__(self, ctx, req): super().__init__(
        verb=ctx.command,
        path=ctx.path,
        head={},
        body=None
    )

    def redirect(self, ctx, location):
        self.reply(ctx, 302, head={'Location': location})

    def respond(self, ctx, status=200, headers={}, message=""):
        if self.__sent:
            raise Exception('Status already sent')
        ctx.send_response(status, message)
        self.__sent = True

        # Finilize the headers
        for key in headers:  # Append headers
            self.head[key] = headers[key]

    def reply(self, ctx, body=None, head={}):
        for key in head:  # Append headers
            self.head[key] = head[key]

        # Send status code (if not sent)
        if not self.__sent:
            self.respond(ctx, self.status)

        # Reply with headers (if not already sent)
        for key in self.head:
            ctx.send_header(key, self.head[key])
        ctx.end_headers()

        # Send the response UTF encoded (if defined)
        if body:
            ctx.wfile.write(body.encode('utf8'))


class Handler(SimpleHTTPRequestHandler):
    # Bind your route handlers into our router's path resolvers
    def __init__(self, server, *extra_args, **kwargs):
        self.reply = server.reply
        self.config = server.config
        super().__init__(directory=server.config.static, *extra_args, **kwargs)

    def do_DEFAULT(self, verb):
        # If no custom routes were triggered, this function will be called..
        # We will check for default actions in this order:
        #  1) config.static - (PATH) Serve static content
        #  2) Fallback: Send "Not found"
        if verb == "GET" and self.config.static:
            # Serve contents from the specified static folder
            return self.do_STATIC()
        else:
            # The default action is to reply: "Not found"
            self.send_response(404, "Not Found")
            self.end_headers()

    def do_REPLY(self, verb): self.reply(verb, self.path, self)
    def do_STATIC(self): super().do_GET()
    def do_HEAD(self): self.do_GET()
    def do_GET(self): self.do_REPLY("GET")
    def do_POST(self): self.do_REPLY("POST")
    def do_PUT(self): self.do_REPLY("PUT")
    def do_PATCH(self): self.do_REPLY("PATCH")
    def do_DELETE(self): self.do_REPLY("DELETE")


class SimpleServer(AbstractServer):

    def __init__(self, prefix='', config=None, app=None):
        # Initialise the base server instance
        super().__init__(prefix, config=config, app=app)

    def start(self):
        # Create a new server instance that will be serving our routes
        self.host = (self.config.host, self.config.port)
        self.app = HTTPServer(self.host, partial(Handler, self))

        # Start the server using the target (request handler) type
        self.onStart()
        self.app.serve_forever()

    def static(self, path):
        # Set the static folder (will be set when server starts)
        self.config.static = path

    def route(self, verb, route):
        print(' + [ %-5s ] %s' % (verb, route))

        def decorator(action):
            # Register this rout, to be searched and resolved in SimpleServer.reply(verb, path, ctx)
            http = self.routes
            http[verb] = {} if not verb in http else http[verb]
            http[verb][route] = action
            return action

        return decorator

    def reply(self, verb, path, ctx):
        # This method is called from the `Handler` interface, each time a route is intercepted
        # See if we can match a route to the current verb
        action = self.find_route(verb, path)
        if not action:
            # Not found: delegate to the default route handler
            return ctx.do_DEFAULT(verb)

        # Service the incomming request with the specified handler
        req = Request(ctx)
        resp = Response(ctx, req)
        data = self.render(action)(req, resp)

        # Write the headers of the response
        ctx.send_response(resp.status)
        for key in resp.head:
            ctx.send_header(key, resp.head[key])
        ctx.end_headers()

        # Send the response UTF encoded (if defined)
        if type(data) == str:
            ctx.wfile.write(data.encode('utf8'))
        elif data:
            raise Exception('Response could not be encoded to text: %s' % data)



import signal

from importlib import import_module

from anyserver import GetConfig
from anyserver.config import PrintConfig
from anyserver.router import WebRouter
from anyserver.templates import TemplateRouter


class AbstractServer(TemplateRouter):
    app = None
    config = None

    def __init__(self, prefix='', config=None, app=None):
        config = config if config else GetConfig()
        super().__init__(prefix, base=config.templates, routes=config.routes)
        self.config = config
        self.app = app

    def start(self):
        raise Exception('Not implemented: BaseServer.start()')

    def register(self, router):
        # Get the raw list of routes, eg: routes[VERB][path] = func(req, resp)
        routes = router._routes() if isinstance(router, WebRouter) else router

        # Update our internal routes
        for verb in routes:
            # Create verb entry if not exist
            for sub_path in routes[verb]:
                # Register the route in this we
                route = self.prefix + sub_path
                action = routes[verb][sub_path]
                self.route(verb, route)(action)

    def route(self, verb, route):
        raise Exception('Not implemented: BaseServer.route(verb, route)')

    def static(self, path):
        raise Exception('Not implemented: BaseServer.static(path)')

    def onStart(self):
        signal.signal(signal.SIGINT, self.onExit)

        # Print server header with config details
        PrintConfig(self.config)

    def onExit(self, signum, frame): return exit(1)

    def discover(self, path="./routes"):
        print(' - Auto discovering routes in: %s' % path)


class OptionalModule:
    _mod = None
    _name = None

    def __init__(self, name, imports=[]):
        try:
            # Try and load the module and the imports
            self._name = name
            self._mod = self.module()
            for ident in imports:
                setattr(self, ident, self.imports(ident))
        except Exception:
            pass  # Failed to load the module and/or some props

    def imports(self, class_name): return getattr(self.module(), class_name)

    def found(self): return not self._mod == None

    def module(self):
        try:
            name = self._name
            self._mod = self._mod if self._mod else import_module(name)
            return self._mod
        except ImportError:
            raise Exception(f"""
# You are trying to use the '{name}' python package dependency. 
# This dependency was not found, and is not currently installed.
# To install `{name}`, run the following command:

> pip3 install -r requirements.txt {name}

WARNING: Flask python package not found. Aborting!
""")

    def hasBase(self, obj, class_name):
        def all_base_classes(type):
            res = [type.__name__]
            for cls in (cls for cls in type.__bases__ if not cls.__name__ == "object"):
                res = res + all_base_classes(cls)
            return res

        return class_name in all_base_classes(obj.__class__)



import logging
import os

from anyserver.router import WebRequest, WebResponse
from anyserver.servers.abstract import AbstractServer, OptionalModule

# Bootstrap the flask module (if available amd installed as a dependency)
uvicorn = OptionalModule('uvicorn', ['run'])
fastapi = OptionalModule('fastapi', [
    'APIRouter',
    'FastAPI',
    'Request',
    'Response'
])
fastapiStatic = OptionalModule('fastapi.staticfiles', ['StaticFiles'])


def tryFastAPIServer(app=None, config=None, prefix=''):
    found = uvicorn.found() and fastapi.found() and fastapiStatic.found()
    if not found or (app and not fastapi.hasBase(app, 'FastAPI')):
        return None  # App instance is not a Flask Application, skip...
    try:
        # Load the flask server if the dependencies are installed
        return FastAPIServer(prefix, config, app)
    except Exception:
        return None


class FastAPIServer(AbstractServer):
    """
    Define a server instance that uses FastAPI ans the underlying engine
    """

    class Request(WebRequest):
        # Wrap your request object into serializable object
        def __init__(self, ctx): super().__init__(
            verb=ctx.method,
            path=ctx.url.path,
            head=ctx.headers,
            body=None,
            params={},
        )

    class Response(WebResponse):
        __sent = False

        # Wrap your response object into serializable object
        def __init__(self, ctx, req):
            super().__init__(
                verb=req.method,
                path=req.url.path,
                head={},
                body=None
            )

        def redirect(self, ctx, location):
            self.status = 302
            self.head['Location'] = location

        def respond(self, ctx, status=200, headers={}, message=""):
            self._resp.status = message
            self._resp.status_code = status
            for key in headers:  # Append headers
                self.head[key] = headers[key]

        def reply(self, ctx, body=None, head={}):
            for key in head:  # Append headers
                self.head[key] = head[key]

            # Send status code (if not sent)
            if not self.__sent:
                self.respond(ctx, self.status)

            # Reply with headers (if not already sent)
            for key in self.head:
                self._resp.headers[key] = self.head[key]

            # Send the response UTF encoded (if defined)
            if body:
                self._resp.data = body

    def __init__(self, prefix='', config=None, app=None):
        self.app = app if app else fastapi.FastAPI()
        super().__init__(prefix, config, self.app)

    def start(self):
        self.onStart()

        # Mount the static path afetr all routes were registered
        static = self.config.static
        if static and os.path.isdir(static):
            fileserver = fastapiStatic.StaticFiles(directory=static, html=True)
            self.app.mount("/", fileserver, name="static")

        # Start the server using the target (request handler) type
        debug = self.config.debug
        imports = self.config.reloads
        if debug and not imports:
            logging.warn('WARNING: Live reload mode has been disabled.')
            logging.warn(' - To use live reload, speficy the app entrypoint.')
            logging.warn('   eg: config["reloads"] = "main:app.app"')
            debug = False

        host = self.config.host
        port = self.config.port
        handle = self.app if not imports else imports
        uvicorn.run(handle, host=host, port=port, reload=debug)

    def static(self, path):
        self.config.static = path # Will be loaded on start

    def route(self, verb, route):
        print(' + [ %-5s ] %s' % (verb, route))


        # Register all routes with the current flask server
        def decorator(action):
            Request = fastapi.Request
            Response = fastapi.Response

            async def respond(request: Request, response: Response):
                async def _body():
                    if not request.method in ["POST", "PUT", "PATCH"]:
                        return None

                    # Parse the body according to the content type
                    ctype = request.headers['content-type'] if 'content-type' in request.headers else None
                    match ctype:
                        case 'application/json':
                            return await request.json()
                        case 'application/x-www-form-urlencoded':
                            return await request.form()

                    return await request.body()

                # Service the incomming request with the specified handler
                # def template(path, data): return render_template(path, **data)
                req = FastAPIServer.Request(request)
                req.body = await _body()  # Fetch the body (async)
                resp = FastAPIServer.Response(response, request)

                data = self.render(action)(req, resp)
                ctype = None if not 'content-type' in resp.head else resp.head['content-type']
                if type(data) == str:
                    return fastapi.Response(content=data, media_type=ctype)

                return data

            # Register the route handler with flask's internal route handling
            # eg: @app.<VERB>(<ROUTE>)
            app = self.app
            router = fastapi.APIRouter()
            router.add_api_route(route, respond, methods=[verb])
            app.include_router(router)

            return action

        return decorator




import io
import csv

from anyserver.encoder import Encoder


class CsvEncoder(Encoder):
    mime = "text/csv"
    ext = [".csv"]

    def encode(self, data):
        data = [data] if not type(data) == list else data
        fields = data[0].keys() if len(data) else []
        with io.StringIO() as buffer:
            writer = csv.DictWriter(buffer, fieldnames=fields)
            writer.writeheader()
            writer.writerows(data)
            return buffer.getvalue()

    def decode(self, data):
        list = []
        with io.StringIO(initial_value=data) as buffer:
            reader = csv.DictReader(buffer)
            for row in reader:
                list.append(row)
        if len(list) < 2:
            return None if len(list) == 0 else list[0]
        return list



import json

from anyserver.encoder import Encoder


class JsonEncoder(Encoder):
    mime = "application/json"
    ext = [".json"]
    def encode(self, data): return json.dumps(data)
    def decode(self, data): return json.loads(data)




import os
import logging
import importlib

from glob import glob

from anyserver.router import WebRouter
from anyserver.encoder import Encoder

logger = logging.getLogger('templates')


class TemplateRouter(WebRouter):
    cache = {}

    def __init__(self, prefix='', base="./templates", routes=None, encoders=None):
        super().__init__(prefix, routes)
        self.base = base
        self.views = {}
        self.default_enc = 'application/json'
        self.encoders = {}
        self.jinja2 = None

        # Map all known encoders for later use
        encoders = encoders if encoders else Encoder.all()
        for enc in encoders:
            self.encoders[enc.mime] = enc

    def renders(self, path, content_type=''):
        if not self.jinja2:
            try:
                # Try and load the jnja templates (if installed)
                self.jinja2 = importlib.import_module('jinja2')
            except ImportError:
                message = "If you use render templates, you need to install jinja2"
                raise Exception(message)

        def decorator(action):
            # Save a ref for this action, linked to the template path
            self.cache[action] = path

            # Now lets try and load all known templates for this path
            self.add_views(path, content_type)
            return self.render(action)  # Render best matching template

        return decorator

    def content_type(self, req, resp, path=None):
        # Try and resolve the template by path and content type
        accept = req.head['accept'] if 'accept' in req.head else ''
        ctype = resp.head['content-type'] if 'content-type' in resp.head else ''
        ctype = '' if ctype == 'application/x-www-form-urlencoded' else ctype

        # Check for edge case where HTMX content is submitted with a form
        if 'hx-request' in req.head:
            return 'text/html'

        # Content type was specified, use as is (dont try and modify)
        if not ctype:
            # Try and find a content type from the accepted content types
            # Accept: text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.7
            views = self.views
            view_types = [] if not path in views else list(views[path].keys())
            for item in [] if not accept else accept.split(','):
                item_type = item.split(';')[0]
                if item_type in view_types:
                    ctype = item_type
                    break
                elif item_type in self.encoders:
                    ctype = item_type
                    break

        return ctype

    def render(self, action):
        # Resolve the template if any were registered
        path = self.cache[action] if action in self.cache else ''

        def wrap_params(req, resp, data):
            return {
                "req": req,
                "resp": resp,
                "request": req,
                "response": resp,
                **data
            }

        def formatted(req, resp, render=None, *args, **kwargs):
            # Get the raw result from the action
            data = action(req, resp, *args, **kwargs)

            # Try and resolve the template by path and content type
            ctype = self.content_type(req, resp, path)
            if ctype:  # Set the response content type (if not already set)
                resp.head['content-type'] = ctype

            # If the result is already a string, we skip, and assume its already encoded
            if data and type(data) == str:
                return data

            filename = self.find_view(path, ctype)
            if not filename:
                # No template, render current encoding from content type
                ctype = self.default_enc if not ctype else ctype
                resp.head['content-type'] = ctype
                return self.encode(data, ctype)

            # Template found, lets render it (incl. request and response objects)
            params = wrap_params(req, resp, data)
            if render:
                # Custom render action was specified, template was found
                return render(filename, params)
            else:
                # Template found, use default template renderer
                return self.render_template(filename, params)

        return formatted

    def render_template(self, path, data):
        base = self.base
        path = path if not path.startswith("/") else path[1:]
        target = os.path.join(base, path)
        with open(target, 'r') as file:
            content = file.read()
            loader = self.jinja2.FileSystemLoader(base)
            template = self.jinja2.Environment(
                loader=loader).from_string(content)
            return template.render(data)

    def add_views(self, path, ctype=None):
        # Create a collection for the view path(s)
        views = self.views
        views[path] = views[path] if path in views else {}

        # Search for known file types (or only the given content type if specified)
        types = [ctype] if ctype else self.encoders.keys()
        for ctype in types:
            found = self.load_view(path, ctype)
            if found:
                views[path][ctype] = found
                logger.debug(' + LOAD[%s] (ctype: %s) <=- %s' %
                             (path, ctype, found))

    def load_view(self, path, ctype):
        # Check if template for content type is cached
        if ctype and ctype in self.views[path]:
            return self.views[path][ctype]

        # Define the file path prefix (excl. extensions)
        base = self.base
        path = path if not path.startswith("/") else path[1:]
        target = os.path.join(base, path)

        # Try and search the filesystem for mappings to the file types
        found = None
        if ctype in self.encoders and len(self.encoders[ctype].ext):
            # Find the extensions for the given mime type
            files = []
            for ext in set(self.encoders[ctype].ext):
                files.extend(glob(target+"*"+ext))
            if len(files):
                found = files[0]  # Return first matched file template

        # Fallback, if the file target exists (no extended prefixes),
        # load it explicitly as the template.
        if not found and os.path.isfile(target):
            found = target

        # Normalise and make path normal
        found = found[len(base):] if found else None
        found = found[1:] if found and found.startswith("/") else found
        return found

    def find_view(self, path, ctype=None):
        logger.debug(' ? FIND[%s] (ctype: %s)' % (path, ctype))
        views = self.views

        # If no content type specified, we will search for available response types
        if path in views and not ctype and len(list(views[path].keys())):
            # Return the default content type, or the first available type (if no default set)
            default = self.default_enc
            available = list(views[path].keys())
            ctype = default if default in views[path] else available[0]
        else:
            # Fall back to default content type (if not specified)
            ctype = self.default_enc if not ctype else ctype

        # Get the render template by path and content type
        if path in views and ctype in views[path]:
            logger.debug(' - GETS[%s] (ctype: %s) - %s' %
                         (path, ctype, views[path][ctype]))
            return views[path][ctype]

        # No render engine found for content type
        return None

    def encode(self, data, ctype=None):
        # Fall back to default content type for encoding
        ctype = self.default_enc if not ctype else ctype

        if ctype not in self.encoders:
            raise Exception('Failed to encode to "%s" content type.' % ctype)

        return self.encoders[ctype].encode(data)



import os
import logging

from anyserver.router import WebRequest, WebResponse
from anyserver.servers.abstract import AbstractServer, OptionalModule

# Bootstrap the flask module (if available amd installed as a dependency)
flask = OptionalModule('flask', [
    'Flask',
    'Response',
    'request',
    'redirect',
    'render_template',
    'send_from_directory'
])


def tryFlaskServer(app=None, config=None, prefix=''):
    if not flask.found() or (app and not flask.hasBase(app, 'Flask')):
        return None  # App instance is not a Flask Application, skip...
    try:
        # Load the flask server if the dependencies are installed
        return FlaskServer(prefix, config, app)
    except Exception:
        return None


class FlaskServer(AbstractServer):
    """
    Define a server instance that uses Flask ans the underlying engine
    """

    class Request(WebRequest):

        # Wrap your request object into serializable object
        def __init__(self, ctx):
            self.ctx = ctx
            self.head = ctx.headers
            self.params = ctx.args
            self.body = self._body(ctx)
            super().__init__(
                verb=ctx.method,
                path=ctx.path,
                head=self.head,
                params=self.params,
                body=self.body
            )

        def _body(self, ctx):
            if not ctx.method in ["POST", "PUT", "PATCH"]:
                return None

            # Parse the body according to the content type
            ctype = ctx.headers['content-type'] if 'content-type' in ctx.headers else None
            match ctype:
                case 'application/json':
                    return ctx.json
                case 'application/x-www-form-urlencoded':
                    return ctx.form

            return ctx.data

    class Response(WebResponse):
        __sent = False

        # Wrap your response object into serializable object
        def __init__(self, ctx, req):
            self.ctx = ctx

            # Bind properties
            super().__init__(
                verb=req.verb,
                path=req.path,
                head={},
                body=None
            )

        def redirect(self, ctx, location):
            flask.redirect(location, 302)

        def respond(self, ctx, status=200, headers={}, message=""):
            self._resp.status = message
            self._resp.status_code = status
            for key in headers:  # Append headers
                self.head[key] = headers[key]

        def reply(self, ctx, body=None, head={}):
            for key in head:  # Append headers
                self.head[key] = head[key]

            # Send status code (if not sent)
            if not self.__sent:
                self.respond(ctx, self.status)

            # Reply with headers (if not already sent)
            for key in self.head:
                self._resp.headers[key] = self.head[key]

            # Send the response UTF encoded (if defined)
            if body:
                self._resp.data = body

    def __init__(self, prefix='', config=None, app=None):
        app = app if app else flask.Flask(__name__)
        super().__init__(prefix, config, app)

    def start(self):
        # Start the server using the target (request handler) type
        debug = self.config.loglevel == logging.DEBUG
        self.onStart()
        self.app.run(debug=debug, host=self.config.host, port=self.config.port)

    def route(self, verb, route):
        print(' + [ %-5s ] %s' % (verb, route))

        # Register all routes with the current flask server
        def decorator(action):
            def respond(*args, **kwargs):

                req = FlaskServer.Request(flask.request)
                resp = FlaskServer.Response(flask.Response(status=200), req)

                # Service the incomming request with the specified handler
                def template(path, data):
                    return flask.render_template(path, **data)

                # Prepare the response and format the result
                response = resp.ctx
                response.data = self.render(action)(req, resp, render=template)
                for head in resp.head.keys():
                    response.headers[head] = resp.head[head]
                return response

            # Register the route handler with flask's internal route handling
            # eg: @self.app.route(route, methods=[verb])(respond)
            self.app.add_url_rule(
                route, route, view_func=respond, methods=[verb])

            return action
        return decorator

    def static(self, path):
        self.config.static = path

        if not os.path.isdir(path):
            return  # Static path does not exist

        # Bind the static content to flask's router
        @self.app.route('/', defaults={'path': 'index.html'})
        @self.app.route('/<path:path>')
        def serve_static(path):
            static = os.path.realpath(self.config.static)
            if not os.path.isfile(static + path) and os.path.exists(static + '/' + path + '/index.html'):
                path = path + 'index.html'
            return flask.send_from_directory(static, path)



import importlib

from anyserver.encoders.base import Encoder
from anyserver.encoders.csv import CsvEncoder
from anyserver.encoders.html import HtmlEncoder
from anyserver.encoders.json import JsonEncoder
from anyserver.encoders.text import TextEncoder
from anyserver.encoders.yaml import YamlEncoder


# Define common encoders
TEXT = TextEncoder()
HTML = HtmlEncoder()
JSON = JsonEncoder()
CSV = CsvEncoder()

# Register common encoders
Encoder.register(
    TEXT,
    HTML,
    CSV,
    JSON
)

# -----------------------------------------------------------------
# Load optional encoders only if the dependencies are installed
# -----------------------------------------------------------------
# Try and load the YAML encoder if `yaml` module found
YAML = None
try:
    YAML = YamlEncoder(importlib.import_module('yaml'))
    Encoder.register(YAML)
except ImportError:
    pass  # Module `yaml` not found...



from anyserver.config import GetConfig
from anyserver.router import WebRouter
from anyserver.encoder import Encoder, CsvEncoder, HtmlEncoder, JsonEncoder, TextEncoder, YamlEncoder

from anyserver.server import AnyServer
from anyserver.servers.abstract import AbstractServer
from anyserver.servers.fastapi import FastAPIServer
from anyserver.servers.flask import FlaskServer
from anyserver.templates import TemplateRouter



from anyserver.servers.simple import SimpleServer
from anyserver.servers.fastapi import tryFastAPIServer
from anyserver.servers.flask import tryFlaskServer

SERVER_TYPES = {
    "FastAPI": tryFastAPIServer,
    "Flask": tryFlaskServer,
}


def AnyServer(app=None, config=None, prefers=None, prefix=''):
    server = None

    # If there is a preferred type, try instantiate that
    if prefers in SERVER_TYPES:
        server = SERVER_TYPES[prefers](app, config, prefix)

    # Now try all of the other registered server types (in the order they were registered)
    for ident in (t for t in SERVER_TYPES if not t == prefers):
        if not server:
            server = SERVER_TYPES[ident](app, config, prefix)
        else:
            break

    # Fall back to the simple server implementation
    if not server:
        server = SimpleServer(prefix, config, app)

    if server:
        print('=' * 64)
        print(f'Starting {server.__class__.__name__}...')
        print('=' * 64)

    return server



import re
import json

from urllib.request import urlopen, Request


class Serializable:
    def toJSON(self):
        return json.dumps(self, default=lambda o: o.__dict__)


class WebRequest(Serializable):
    def __init__(self, verb, path, head={}, body=None, params={}):
        self.verb = verb
        self.path = path
        self.head = head
        self.body = body
        self.params = params


class WebResponse(Serializable):
    def __init__(self, verb, path, head, body, status=200):
        self.verb = verb
        self.path = path
        self.status = status
        self.head = head
        self.body = body


class WebRouter:

    def __init__(self, prefix='', routes=None):
        self.prefix = prefix
        self.routes = routes if routes else {}

    def register(self, router):
        routes = router  # If directly imported

        # Check if routes were registered into a router
        if isinstance(router, WebRouter):
            # Web router was used to register routes
            routes = router._routes()

        # Update our internal routes
        http = self.routes
        for verb in routes:
            # Create verb entry if not exist
            if not verb in http:
                http[verb] = {}
            for sub_path in routes[verb]:
                # Register the route in this we
                route = self.prefix + sub_path
                action = routes[verb][sub_path]
                http[verb][route] = action
                print(' + [ %-5s ] %s' % (verb, route))

    def _routes(self):
        # Get the parse list of routes that are registered
        http = {}
        prefix = self.prefix if self.prefix != '/' else ''
        routes = self.routes
        for verb in routes:
            http[verb] = {} if not verb in http else http[verb]

            # Create verb entry if not exist
            for sub_path in routes[verb]:
                action = routes[verb][sub_path]
                route = prefix + sub_path
                http[verb][route] = action

        return http

    def find_route(self, verb, path):
        # Search locally registered routes for a route handler
        if not self.routes or not verb in self.routes:
            return None
        routes = list(self.routes[verb].keys())
        routes.sort(reverse=True, key=len)
        matched = [r for r in routes if re.search(r, path)]
        if len(matched) > 0:
            route = matched[0]  # Return the first match
            return self.routes[verb][route]
        else:
            return None

    def default(self, verb, path):
        # This method should be extended by a server implementation
        message = "You need to implement the `default(self, verb, path)` function.\n"
        message += "Verb: %s, Path: %s\n" % (verb, path)
        raise Exception('FATAL: %s' % message)

    def route(self, verb, path):
        routes = self.routes
        if not verb in routes:
            routes[verb] = {}

        def decorator(action):
            routes[verb][path] = action
            return action

        return decorator

    def head(self, path):
        return self.route("HEAD", path)

    def get(self, path):
        return self.route("GET", path)

    def post(self, path):
        return self.route("POST", path)

    def put(self, path):
        return self.route("PUT", path)

    def patch(self, path):
        return self.route("PATCH", path)

    def delete(self, path):
        return self.route("DELETE", path)

    def proxy(self, url, req):
        if not url:
            raise Exception("FATAL: No proxy URL has been set up.")

        url = '{}{}'.format(url, req.path)
        print(' ~ Proxy me: %s' % url)

        # Populate the new request with the headers that was requested from client
        headers = {}
        for key in req.head:
            name = key.lower()
            value = req.head[key]
            if name == "host":  # <-- Trick endpoint into thinking its a direct call
                proxy_host = url.replace('http://', '')
                proxy_host = proxy_host.replace('https://', '')
                proxy_host = proxy_host.replace('localhost', '127.0.0.1')
                proxy_host = proxy_host.split('/')[0]
                value = proxy_host
                pass
            if name.startswith('x-') or name.startswith('sec-') or name in (
                'connection',
                'user-agent'
            ):
                pass  # <-- Filtering out noise and tracers we dont need
            else:
                headers[key] = value

        # Create a new request handler, then fetch the response via a proxied request
        req = Request(url, headers=headers)
        resp = urlopen(req)
        return resp





from anyserver.encoder import Encoder


class TextEncoder(Encoder):
    mime = "text/plain"
    ext = [".txt"]
    def encode(self, data): return '{}'.format(data)
    def decode(self, data): return '{}'.format(data)




import os
import sys
import logging
import argparse

from anyserver.encoder import Encoder


class ServerConfig:
    # Host configuration
    host = '0.0.0.0'
    port = 9999

    # Runtime settings
    working = "."
    templates = "./templates"
    plugins = []

    # Optional: Define the default route hander as one of the following
    static = None
    proxy = None

    debug = False
    discover = True
    reloads = None
    routes = {}


def GetConfig():
    opts = GetArgs()
    
    config = ServerConfig()

    # Try and load from config file (if specified)
    if opts.config and os.path.isfile(opts.config):
        configDict = Encoder.loadFile(opts.config)
        for key in configDict.keys():
            setattr(config, key, configDict[key])

    # Apply and load additional config settings
    ApplyArgs(config, opts)  # <-- Apply CLI args to config

    # Apply any CLI args and return the config
    return config


def GetArgs(argv=sys.argv[1:]):
    parser = argparse.ArgumentParser(
        description='Extremely simple python server that can be easily be extended.'
    )

    # Core configurations and valiables used by all server types
    parser.add_argument('-c', '--config',
                        dest='config',
                        help='Specify your `config.yaml` file.',
                        )
    parser.add_argument('--host',
                        dest='host',
                        help='Specify the host we will bind the server to',
                        default='0.0.0.0'
                        )
    parser.add_argument('-p', '--port',
                        dest='port',
                        help='serve HTTP requests on specified port (default: 9999)',
                        type=int,
                        default=9999
                        )
    parser.add_argument('-w', '--working',
                        dest='working',
                        help='Specify the working directory',
                        default=os.getenv('WORK_DIR')
                        )

    # Allow the loading of plugins
    parser.add_argument('-a', '--add',
                        dest='plugins',
                        help='List of plugins to load',
                        type=lambda s: [item for item in s.split(',')])

    # Allow the user to specify the type of server that will be created
    parser.add_argument('-s', '--static',
                        dest='static',
                        help='Static web contents to serve, if no other route defined',
                        # Default option: Serve current dir...
                        default=os.getenv('STATIC_DIR', '')
                        )
    parser.add_argument('--proxy',
                        dest='proxy',
                        help='Default route handler will reverse proxy to a URL'
                        )

    parser.add_argument('--debug',
                        help="Print debugging statements",
                        dest="debug",
                        action="store_const",
                        const=True,
                        default=False,
                        )
    parser.add_argument('--verbose',
                        help="Be verbose",
                        action="store_const", dest="loglevel",
                        const=logging.INFO,
                        default=logging.WARN,
                        )

    # Parse the args provided by the CLI
    args = parser.parse_args(argv)

    # Set the log verbosity
    format = '%(message)s'
    loglevel = logging.DEBUG if args.debug else args.loglevel
    logging.basicConfig(format=format, level=loglevel)

    return args


def ApplyArgs(config, args):
    for key, val in vars(args).items():
        if val and key != "config":
            setattr(config, key, val)
    return config


def printIf(message, value=None):
    if value:
        print(message % value)


def PrintConfig(config):
    print('-' * 64)
    printIf(' + Work Dir: %s', config.working)
    printIf(' + WWW Root: %s', config.static)
    printIf(' ~ Proxy To: %s', config.proxy)
    print(' - Hostname: http://%s:%s' % (config.host, config.port))
    print('-' * 64)




from anyserver.encoder import Encoder


class HtmlEncoder(Encoder):
    mime = "text/html"
    ext = [".htm", ".html", ".htmx"]
    def encode(self, data): return '<head></head><body>{}</body>'.format(data)
    def decode(self, data): return '{}'.format(data)




import subprocess
from distutils.core import setup

name = 'anyserver'
version = '0.1.8'

print(f'SETUP {name} [{version}]')
setup(
    name=name,
    packages=[
        name,
        f'{name}.encoders',
        f'{name}.servers',
    ],
    version=version,
    license='MIT',
    description='Simple and generic python web server and utils',
    author='Club404',
    author_email='info@club404.io',
    url='https://github.com/JohnnyBeProgramming/anyserver.py',
    download_url='https://github.com/JohnnyBeProgramming/anyserver.py/archive/refs/tags/%s.tar.gz' % version,
    keywords=['simple', 'http', 'server'],
    install_requires=[],
    classifiers=[  # Optional
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',

        # Pick your license as you wish
        'License :: OSI Approved :: MIT License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
)



REGISTERED = []


class Encoder:
    mime = None
    ext = []

    def encode(self, data):
        raise Exception('Not implemented: Encoder.encode()')

    def decode(self, data):
        raise Exception('Not implemented: Encoder.decode()')

    @staticmethod
    def all():
        return REGISTERED  # Return all registered

    @staticmethod
    def register(*encoders):
        for enc in filter(lambda enc: not enc in REGISTERED, encoders):
            REGISTERED.append(enc)

    @staticmethod
    def remove(*encoders):
        for enc in filter(lambda enc: enc in REGISTERED, encoders):
            REGISTERED.remove(enc)

    @staticmethod
    def find(filename):
        # Determine file type from extension
        parts = filename.split('.')
        type = None if len(parts) < 2 else '.'+parts[-1]

        # Find the first registered encoder that can handle this file type
        encoder = next(filter(lambda enc: type in enc.ext, REGISTERED), None)

        if not encoder:
            raise Exception(f"Cannot decode '{filename}' (file type: {type}).")

        return encoder

    @staticmethod
    def loadFile(filename):
        # Read the file contents and decode with the specified encoding
        encoder = Encoder.find(filename)
        with open(filename, 'r') as f:
            return encoder.decode(f.read())

    @staticmethod
    def saveFile(filename):
        # Read the file contents and decode with the specified encoding
        encoder = Encoder.find(filename)
        with open(filename, 'r') as f:
            return encoder.decode(f.read())
